_ = require "underscore"
fs = require "fs"
path = require "path"

# FIXME: This syntax will be deprecated, needs to be replaced
#require.paths.unshift "#{process.cwd()}/node_modules"

# Catch-all error handling
process.on "uncaughtException", (e) ->
  # FIXME: This is a temporary hack until the `cradle` lib gets patched
  # so that errors can be properly handled directly by dk-couchdb
  if e.message == "ECONNREFUSED, Connection refused"
    console.log "ERROR: Unable to connect to CouchDB, is it running?"
    process.exit()
  else
    console.log e.toString()

dk = {}
runners = []
plugins = []
loadedPlugins = {}
pluginExtensions = {}
commands = {}

dk.boot = (options={}) ->
  dk.env = options.env || "development"
  dk.loadModules()
  dk.config()
  dk.run()

dk.commands = -> commands

dk.onRun = (f) ->
  runners.push f

dk.run = (options) ->
  runners.forEach (runner) -> runner()

dk.runCommand = (command, options={}) ->
  dk.loadModules()
  meta = commands[command]
  unless meta
    throw "DrumKit is missing the command '#{command}'"

  # FIXME: Circular dependencies need to be handled
  for dep in meta.dependencies
    dk.loadPlugin dep

  meta.init options

dk.command = (name, summary, deps, func) ->
  if _.isFunction deps
    func = deps
    deps = []
  commands[name] = {summary: summary, dependencies: deps, init: func}

dk.plugin = (namespace, deps, func) ->
  if _.isFunction deps
    func = deps
    deps = []
  plugins[namespace] = {dependencies: deps, init: func}

dk.extendPlugin = (namespace, deps, func) ->
  if _.isFunction deps
    func = deps
    deps = []
  pluginExtensions[namespace] ?= []
  pluginExtensions[namespace].push {dependencies: deps, init: func}

dk.loadPlugin = (namespace) ->
  return if loadedPlugins[namespace]
  loadedPlugins[namespace] = true

  meta = plugins[namespace]
  return unless meta

  # FIXME: Circular dependencies need to be handled
  for dep in meta.dependencies
    dk.loadPlugin dep

  parts = namespace.split "."
  moduleName = parts.pop()
  currentNs = dk
  currentNsName = null
  for part in parts
    if currentNsName
      currentNsName += ".#{part}"
    else
      currentNsName = part
    dk.loadPlugin currentNsName

    currentNs[part] ?= {}
    currentNs = currentNs[part]

  mod = meta.init()
  
  if pluginExtensions[namespace]
    pluginExtensions[namespace].forEach (extMeta) ->
      extMeta.dependencies.forEach (p) -> dk.loadPlugin(p)
      mod = extMeta.init(mod)

  currentNs[moduleName] = mod
  if mod.run
    runners.push mod.run

loadPlugins = ->
  for namespace, func of plugins
    dk.loadPlugin namespace

dk.loadModules = ->
  if path.existsSync("package.json")
    data = JSON.parse fs.readFileSync("package.json")
    _.each _.keys(data.dependencies), (name) ->
      if /^dk/.test name
        try
          require name
        catch e
          require "#{process.cwd()}/node_modules/#{name}"
    loadPlugins()
  else
    # console.log "No package.json found, no npm modules autoloaded."
  loadAppPluginModules()

loadAppPluginModules = ->
  pluginsDir = "#{process.cwd()}/plugins"
  if path.existsSync pluginsDir
    for dir in fs.readdirSync pluginsDir
      console.log "Requiring custom plugin ", dir
      require "#{pluginsDir}/#{dir}"

dk.config = ->
  try
    require "#{process.cwd()}/config"
  catch e
    console.log "Unable to find config file, continuing without it..."
  
global.dk = dk
module.exports = dk
